@@ -87,67 +87,6 @@ static unsigned bkey_get_dev_refs(struct bch_fs *c, struct bkey_s_c k)
 	return ptrs_held;
 }
 
-noinline_for_stack
-static void count_data_update_key_fail(struct data_update *u,
-				       struct bkey_s_c new,
-				       struct bkey_s_c wrote,
-				       struct bkey_i *insert,
-				       const char *msg)
-{
-	struct bch_fs *c = u->op.c;
-
-	if (u->stats) {
-		atomic64_inc(&u->stats->keys_raced);
-		atomic64_add(insert->k.size, &u->stats->sectors_raced);
-	}
-
-	event_add_trace(c, data_update_key_fail, insert->k.size, buf, ({
-		prt_str(&buf, bch2_data_update_type_strs[u->opts.type]);
-		prt_newline(&buf);
-
-		prt_str(&buf, msg);
-		prt_newline(&buf);
-
-		struct bkey_s_c old = bkey_i_to_s_c(u->k.k);
-		unsigned rewrites_found = 0;
-
-		if (insert) {
-			const union bch_extent_entry *entry;
-			struct bch_extent_ptr *ptr;
-			struct extent_ptr_decoded p;
-
-			unsigned ptr_bit = 1;
-			bkey_for_each_ptr_decode(old.k, bch2_bkey_ptrs_c(old), p, entry) {
-				if ((ptr_bit & u->opts.ptrs_kill ) &&
-				    (ptr = bch2_extent_has_ptr(c, old, p, bkey_i_to_s(insert))) &&
-				    !ptr->cached)
-					rewrites_found |= ptr_bit;
-				ptr_bit <<= 1;
-			}
-		}
-
-		prt_str(&buf, "rewrites found:\t");
-		bch2_prt_u64_base2(&buf, rewrites_found);
-		prt_newline(&buf);
-
-		bch2_data_update_opts_to_text(&buf, c, &u->op.opts, &u->opts);
-
-		prt_str(&buf, "\nold:    ");
-		bch2_bkey_val_to_text(&buf, c, old);
-
-		prt_str(&buf, "\nnew:    ");
-		bch2_bkey_val_to_text(&buf, c, new);
-
-		prt_str(&buf, "\nwrote:  ");
-		bch2_bkey_val_to_text(&buf, c, wrote);
-
-		if (insert) {
-			prt_str(&buf, "\ninsert: ");
-			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(insert));
-		}
-	}));
-}
-
 static unsigned ptr_remap(struct bch_fs *c, struct bkey_s_c old,
 			  struct extent_ptr_decoded oldp,
 			  struct bkey_s_c new)
@@ -209,6 +148,55 @@ static unsigned bkey_ptr_conflicts_mask(struct bch_fs *c, struct bkey_s_c k1, st
 	return ptrs_conflict;
 }
 
+static void data_update_key_to_text(struct printbuf *out,
+				    struct data_update *u,
+				    struct bkey_s_c new,
+				    struct bkey_s_c wrote,
+				    struct bkey_i *insert)
+{
+	struct bch_fs *c = u->op.c;
+	struct bkey_s_c old = bkey_i_to_s_c(u->k.k);
+
+	prt_str(out, "rewrites found:\t");
+	bch2_prt_u64_base2(out,
+			   ptr_mask_remap(c, old, u->opts.ptrs_kill, bkey_i_to_s_c(insert)));
+	prt_newline(out);
+
+	bch2_data_update_opts_to_text(out, c, &u->op.opts, &u->opts);
+
+	prt_str(out, "\nold:    ");
+	bch2_bkey_val_to_text(out, c, old);
+
+	prt_str(out, "\nnew:    ");
+	bch2_bkey_val_to_text(out, c, new);
+
+	prt_str(out, "\nwrote:  ");
+	bch2_bkey_val_to_text(out, c, wrote);
+
+	prt_str(out, "\ninsert: ");
+	bch2_bkey_val_to_text(out, c, bkey_i_to_s_c(insert));
+}
+
+noinline_for_stack
+static void count_data_update_key_fail(struct data_update *u,
+				       struct bkey_s_c new,
+				       struct bkey_s_c wrote,
+				       struct bkey_i *insert,
+				       const char *msg)
+{
+	struct bch_fs *c = u->op.c;
+
+	if (u->stats) {
+		atomic64_inc(&u->stats->keys_raced);
+		atomic64_add(insert->k.size, &u->stats->sectors_raced);
+	}
+
+	event_add_trace(c, data_update_key_fail, insert->k.size, buf, ({
+		prt_printf(&buf, "%s\n", msg);
+		data_update_key_to_text(&buf, u, new, wrote, insert);
+	}));
+}
+
 static int data_update_index_update_key(struct btree_trans *trans,
 					struct data_update *u,
 					struct btree_iter *iter)
@@ -247,18 +235,37 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	bch2_cut_back(new->k.p,		insert);
 	bch2_cut_back(insert->k.p,	&new->k_i);
 
+	struct bpos next_pos = insert->k.p;
+
 	if (!bch2_extents_match(c, k, old)) {
 		count_data_update_key_fail(u, k, bkey_i_to_s_c(&new->k_i), insert, "no match:");
-		bch2_btree_iter_advance(iter);
+		bch2_btree_iter_set_pos(iter, next_pos);
 		return 0;
 	}
 
-	if (!u->opts.no_devs_have) {
+	unsigned ptrs_kill = ptr_mask_remap(c, old, u->opts.ptrs_kill,	bkey_i_to_s_c(insert));
+	if (ptrs_kill) {
 		/* Don't replace non-cached replicas with cached: */
-		if (u->opts.ptrs_kill)
-			bkey_for_each_ptr(bch2_bkey_ptrs_c(bkey_i_to_s_c(&new->k_i)), ptr)
-				BUG_ON(ptr->cached);
+		bool replacing_non_cached_replicas = false;
+		unsigned ptr_bit = 1;
+		bkey_for_each_ptr(bch2_bkey_ptrs_c(bkey_i_to_s_c(insert)), ptr) {
+			if (ptrs_kill & ptr_bit)
+				replacing_non_cached_replicas |= !ptr->cached;
+			ptr_bit <<= 1;
+		}
 
+		if (replacing_non_cached_replicas)
+			bkey_for_each_ptr(bch2_bkey_ptrs_c(bkey_i_to_s_c(&new->k_i)), ptr)
+				if (ptr->cached) {
+					CLASS(bch_log_msg_ratelimited, msg)(c);
+					prt_printf(&msg.m, "data update tried to replace non cached data with cached:\n");
+					data_update_key_to_text(&msg.m, u, k, bkey_i_to_s_c(&new->k_i),
+								insert);
+					return 0;
+				}
+	}
+
+	if (!u->opts.no_devs_have) {
 		/*
 		 * Drop replicas in the existing extent that conflict with what
 		 * we just wrote - but only if they were explicitly specified in
@@ -325,6 +332,24 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	struct bch_inode_opts opts;
 	try(bch2_bkey_get_io_opts(trans, NULL, k, &opts));
 
+	struct bkey_durability old_durability, new_durability;
+	try(bch2_bkey_durability(trans, k, &old_durability));
+	try(bch2_bkey_durability(trans, bkey_i_to_s_c(insert), &new_durability));
+
+	if (new_durability.total < old_durability.total &&
+	    new_durability.total < min(u->op.opts.data_replicas, opts.data_replicas)) {
+		/*
+		 * This can happen when a move - evacuate or copygc - races with an
+		 * extent being erasure coded: we replaced a pointer with one on the
+		 * same device (so we have to drop it early, to avoid conflicts) which
+		 * is no longer erasure coded.
+		 *
+		 * XXX: trace this
+		 */
+		bch2_btree_iter_set_pos(iter, next_pos);
+		return 0;
+	}
+
 	try(bch2_bkey_drop_extra_durability(trans, &opts,	insert,
 		ptr_mask_remap(c, old, u->opts.ptrs_io_error,	bkey_i_to_s_c(insert)), true));
 
@@ -368,12 +393,9 @@ static int data_update_index_update_key(struct btree_trans *trans,
 					!should_check_enospc
 					? BCH_DISK_RESERVATION_NOFAIL : 0));
 
-	struct bpos next_pos = insert->k.p;
-
 	try(bch2_trans_log_str(trans, bch2_data_update_type_strs[u->opts.type]));
 	try(bch2_trans_log_bkey(trans, u->btree_id, 0, u->k.k));
 
-	try(bch2_insert_snapshot_whiteouts(trans, u->btree_id, k.k->p, bkey_start_pos(&insert->k)));
 	try(bch2_insert_snapshot_whiteouts(trans, u->btree_id, k.k->p, insert->k.p));
 
 	/*
@@ -404,23 +426,7 @@ static int data_update_index_update_key(struct btree_trans *trans,
 				   old_durability.total,
 				   new_durability.total,
 				   opts.data_replicas);
-			prt_str(&msg.m, "orig: ");
-			bch2_bkey_val_to_text(&msg.m, c, k);
-			prt_newline(&msg.m);
-
-			prt_str(&msg.m, "wrote: ");
-			bch2_bkey_val_to_text(&msg.m, c, bkey_i_to_s_c(bch2_keylist_front(&u->op.insert_keys)));
-			prt_newline(&msg.m);
-
-			prt_str(&msg.m, "new: ");
-			bch2_bkey_val_to_text(&msg.m, c, bkey_i_to_s_c(&new->k_i));
-			prt_newline(&msg.m);
-
-			prt_str(&msg.m, "insert: ");
-			bch2_bkey_val_to_text(&msg.m, c, bkey_i_to_s_c(insert));
-			prt_newline(&msg.m);
-
-			bch2_data_update_to_text(&msg.m, u);
+			data_update_key_to_text(&msg.m, u, k, bkey_i_to_s_c(&new->k_i), insert);
 			bch2_fs_emergency_read_only(c, &msg.m);
 			return bch_err_throw(c, emergency_ro);
 		}
@@ -986,6 +992,9 @@ static int bch2_can_do_write_btree(struct bch_fs *c,
 	struct bch_devs_list empty = {};
 	bool need_copygc = false;
 
+	if (bch2_bkey_nr_dirty_ptrs(c, k) > opts->data_replicas)
+		return 0;
+
 	if (durability_available_on_target(c, watermark, BCH_DATA_btree, data_opts->target, &empty,
 					   data_opts->write_flags & BCH_WRITE_alloc_nowait,
 					   trace, &need_copygc) >
