@@ -665,6 +665,12 @@ int bch2_dev_remove(struct bch_fs *c, struct bch_dev *ca, int flags,
 	if (ret)
 		goto err;
 
+	/*
+	 * Disallow reads before we remove alloc info, otherwise we'll get
+	 * spurious stale pointer errors:
+	 */
+	__bch2_dev_offline(c, ca);
+
 	ret = bch2_dev_remove_alloc(c, ca);
 	if (ret) {
 		prt_printf(err, "bch2_dev_remove_alloc() error: %s\n", bch2_err_str(ret));
@@ -713,8 +719,6 @@ int bch2_dev_remove(struct bch_fs *c, struct bch_dev *ca, int flags,
 		goto err;
 	}
 
-	__bch2_dev_offline(c, ca);
-
 	scoped_guard(mutex, &c->sb_lock)
 		rcu_assign_pointer(c->devs[ca->dev_idx], NULL);
 
@@ -995,7 +999,7 @@ static int bch2_dev_may_offline(struct bch_fs *c, struct bch_dev *ca, int flags,
 	struct bch_devs_mask new_rw_devs = c->allocator.rw_devs[0];
 	__clear_bit(ca->dev_idx, new_devs.d);
 
-	if (!bch2_can_read_fs_with_devs(c, new_devs, flags, err) ||
+	if (!bch2_can_read_fs_with_devs(c, &new_devs, flags, err) ||
 	    (!c->opts.read_only &&
 	     !bch2_can_write_fs_with_devs(c, new_rw_devs, flags, err))) {
 		prt_printf(err, "Cannot offline required disk\n");
