@@ -127,6 +127,7 @@ enum bch_fsck_flags {
 	x(alloc_key_stripe_wrong,				110,	FSCK_AUTOFIX)	\
 	x(alloc_key_stripe_redundancy_wrong,			111,	FSCK_AUTOFIX)	\
 	x(alloc_key_journal_seq_in_future,			298,	FSCK_AUTOFIX)	\
+	x(alloc_key_stripe_refcount_wrong,			345,	FSCK_AUTOFIX)	\
 	x(bucket_sector_count_overflow,				112,	0)		\
 	x(bucket_metadata_type_mismatch,			113,	0)		\
 	x(need_discard_key_wrong,				114,	FSCK_AUTOFIX)	\
@@ -160,7 +161,8 @@ enum bch_fsck_flags {
 	x(extent_ptrs_redundant_stripe,				139,	0)		\
 	x(extent_ptrs_unwritten,				140,	0)		\
 	x(extent_ptrs_written_and_unwritten,			141,	0)		\
-	x(extent_ptrs_all_invalid,				338,	0)		\
+	x(extent_ptrs_all_invalid,				338,	FSCK_AUTOFIX)	\
+	x(extent_ptrs_all_invalid_but_cached,			351,	FSCK_AUTOFIX)	\
 	x(extent_reconcile_bad_pending,				332,	0)		\
 	x(extent_reconcile_bad_hipri,				333,	0)		\
 	x(extent_reconcile_bad_replicas,			339,	0)		\
@@ -200,6 +202,9 @@ enum bch_fsck_flags {
 	x(stripe_csum_granularity_bad,				290,	0)		\
 	x(stripe_sectors_zero,					340,	0)		\
 	x(stripe_sector_count_wrong,				169,	0)		\
+	x(stripe_to_missing_bucket_ref,				346,	FSCK_AUTOFIX)	\
+	x(bucket_stripe_ref_to_missing_stripe,			347,	FSCK_AUTOFIX)	\
+	x(bucket_stripe_ref_to_incorrect_stripe,		348,	FSCK_AUTOFIX)	\
 	x(snapshot_tree_pos_bad,				170,	0)		\
 	x(snapshot_tree_to_missing_snapshot,			171,	0)		\
 	x(snapshot_tree_to_missing_subvol,			172,	0)		\
@@ -234,6 +239,7 @@ enum bch_fsck_flags {
 	x(inode_unlinked_but_nlink_nonzero,			198,	0)		\
 	x(inode_unlinked_and_not_open,				281,	0)		\
 	x(inode_unlinked_but_has_dirent,			285,	0)		\
+	x(unlinked_inode_not_on_deleted_list,			244,	FSCK_AUTOFIX)	\
 	x(inode_checksum_type_invalid,				199,	0)		\
 	x(inode_compression_type_invalid,			200,	0)		\
 	x(inode_subvol_root_but_not_dir,			201,	0)		\
@@ -293,7 +299,6 @@ enum bch_fsck_flags {
 	x(dir_loop,						241,	FSCK_AUTOFIX)	\
 	x(hash_table_key_duplicate,				242,	FSCK_AUTOFIX)	\
 	x(hash_table_key_wrong_offset,				243,	FSCK_AUTOFIX)	\
-	x(unlinked_inode_not_on_deleted_list,			244,	FSCK_AUTOFIX)	\
 	x(reflink_p_front_pad_bad,				245,	0)		\
 	x(journal_entry_dup_same_device,			246,	0)		\
 	x(inode_bi_subvol_missing,				247,	0)		\
@@ -315,7 +320,7 @@ enum bch_fsck_flags {
 	x(btree_ptr_v2_min_key_bad,				262,	0)		\
 	x(btree_root_unreadable_and_scan_found_nothing,		263,	0)		\
 	x(snapshot_node_missing,				264,	FSCK_AUTOFIX)	\
-	x(dup_backpointer_to_bad_csum_extent,			265,	0)		\
+	x(dup_backpointer_to_bad_csum_extent,			265,	FSCK_AUTOFIX)	\
 	x(btree_bitmap_not_marked,				266,	FSCK_AUTOFIX)	\
 	x(sb_clean_entry_overrun,				267,	0)		\
 	x(btree_ptr_v2_written_0,				268,	0)		\
@@ -340,6 +345,8 @@ enum bch_fsck_flags {
 	x(logged_op_but_clean,					283,	FSCK_AUTOFIX)	\
 	x(compression_opt_not_marked_in_sb,			295,	FSCK_AUTOFIX)	\
 	x(compression_type_not_marked_in_sb,			296,	FSCK_AUTOFIX)	\
+	x(compression_error,					342,	0)		\
+	x(data_update_got_non_extent,				350,	0)		\
 	x(directory_size_mismatch,				303,	FSCK_AUTOFIX)	\
 	x(dirent_cf_name_too_big,				304,	0)		\
 	x(dirent_stray_data_after_cf_name,			305,	0)		\
@@ -353,7 +360,10 @@ enum bch_fsck_flags {
 	x(btree_ptr_with_no_reconcile_bp,			335,	FSCK_AUTOFIX)	\
 	x(btree_ptr_with_bad_reconcile_bp,			336,	FSCK_AUTOFIX)	\
 	x(btree_ptr_to_bad_reconcile_bp,			337,	FSCK_AUTOFIX)	\
-	x(MAX,							342,	0)
+	x(vfs_i_size_bad,					343,	0)		\
+	x(vfs_i_sectors_bad,					344,	0)		\
+	x(vfs_unlink_got_wrong_inum,				349,	0)		\
+	x(MAX,							352,	0)
 
 enum bch_sb_error_id {
 #define x(t, n, ...) BCH_FSCK_ERR_##t = n,
